import logging
import queue
import threading
from typing import Optional
from uuid import uuid4, UUID

from .Exceptions import UnhandledMessageError, LoginFailedError, InterruptClient
from ..ai import AI
from ..communication import Connection, JchessMessage, JchessMessageType
from ..communication.schema import LoginMessage

_logger = logging.getLogger("j_chess_lib")


class Client(threading.Thread):

    def __init__(self, connection: Connection, ai: AI):
        self._id = uuid4()
        self._original_id = self._id
        self._ai = ai
        self._connection = connection
        self._send_queue: queue.Queue[JchessMessage] = queue.Queue()
        self._recv_queue: queue.Queue[JchessMessage] = queue.Queue()
        super().__init__(daemon=True, name=f"ClientThread-{ai.name}")

        class SendThread(threading.Thread):
            def __init__(self, _connection: Connection, _send_queue: queue.Queue[JchessMessage]):
                super().__init__(daemon=True)
                self._connection = _connection
                self._send_queue = _send_queue

            def run(self) -> None:
                while True:
                    message = self._send_queue.get()
                    self._connection.send(message=message)

        class RecvThread(threading.Thread):
            def __init__(self, _connection: Connection, _recv_queue: queue.Queue[JchessMessage]):
                super().__init__(daemon=True)
                self._connection = _connection
                self._recv_queue = _recv_queue

            def run(self) -> None:
                while True:
                    message = self._connection.recv()
                    # _logger.info("MESSAGE", message)
                    self._recv_queue.put(message)

        send_thread = SendThread(_connection=connection, _send_queue=self._send_queue)
        recv_thread = RecvThread(_connection=connection, _recv_queue=self._recv_queue)
        send_thread.start()
        recv_thread.start()

    @property
    def id(self) -> UUID:
        return self._id

    @property
    def ai(self) -> AI:
        return self._ai

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):
        if isinstance(other, Client):
            return other.id == self.id
        return False

    def __str__(self):
        return f"Client {self.id} with {self.ai.name}"

    def base_msg(self) -> JchessMessage:
        return JchessMessage(player_id=str(self.id))

    def _send(self, message: JchessMessage):
        self._send_queue.put(message)

    def _recv(self) -> JchessMessage:
        message = self._recv_queue.get()
        message = self._intercept_disconnect(message=message)
        message = self._intercept_heartbeat(message=message)
        if message is None:
            return self._recv()
        return message

    def run(self):
        try:
            # --Region Login--
            new_id = self._handle_login()
            if new_id is None:
                _logger.info("Login failed")
                raise LoginFailedError()
            _logger.info(f"Login success{f' id changed {self._id} -> {new_id}'}" if self._id != new_id else '',
                         extra={"AI": self._ai, "Connection": self._connection})
            self._original_id = self._id
            self._id = new_id
            # --Region Login--

            while True:
                message = self._recv()
                if message.message_type == JchessMessageType.MATCH_FOUND:
                    self._handle_match(message=message)

        except InterruptClient as e:
            _logger.error("Client interrupeted and killed", extra={"AI": self._ai, "Connection": self._connection})
            _logger.error(f"Cause: {e}", extra={"AI": self._ai, "Connection": self._connection})

    def _intercept_disconnect(self, message: JchessMessage) -> Optional[JchessMessage]:
        if message is None:
            return message
        if message.message_type == JchessMessageType.DISCONNECT:
            raise InterruptClient(message.disconnect.error_type_code)
        return message

    def _intercept_heartbeat(self, message: JchessMessage) -> Optional[JchessMessage]:
        if message is None:
            return message
        if message.message_type == JchessMessageType.HEART_BEAT:
            _logger.debug("Received a heartbeat", extra={"AI": self._ai, "Connection": self._connection})
            return None
        return message

    def _handle_login(self) -> Optional[UUID]:
        login = LoginMessage(name=self.ai.name)
        message = self.base_msg()
        message.login = login
        message.message_type = JchessMessageType.LOGIN
        self._send(message=message)
        message = self._recv()
        if message.message_type == JchessMessageType.LOGIN_REPLY:
            repl_id = message.login_reply.new_id
            return UUID(repl_id)
        elif message.message_type == JchessMessageType.ACCEPT:
            return self._handle_login()
        elif message.message_type == JchessMessageType.DISCONNECT:
            return None
        else:
            raise UnhandledMessageError(message=message)

    def _handle_match(self, message: JchessMessage):
        from .match import Match
        match: Match = Match.handle_match(message=message, recv=self._recv, ai=self.ai, client=self, send=self._send)
        match.play_match()



if __name__ == "__main__":
    import logging

    logger = logging.getLogger("j_chess_lib")
    logger.setLevel(logging.DEBUG)

    from j_chess_lib.communication import Connection
    from j_chess_lib.client import Client
    from j_chess_lib.ai.examples import Random

    with Connection("localhost", 5123) as connection1:
        with Connection("localhost", 5123) as connection2:
            ai1 = Random("RNJesus", min_turn_time=1)
            ai2 = Random(min_turn_time=1)

            client1 = Client(connection=connection1, ai=ai1)
            client2 = Client(connection=connection2, ai=ai2)

            client1.start()
            logger.info("Client1 started")
            client2.start()
            logger.info("Client2 started")

            client1.join()
            logger.info("Client1 finished")
            client2.join()
            logger.info("Client2 finished")



"""Top-level package for j-chess-lib."""
import logging
import sys

logger = logging.getLogger("j_chess_lib")
logger_handler = logging.StreamHandler(stream=sys.stderr)
logger_formatter = logging.Formatter('{asctime} - {levelname:^8s} - {message}', "%Y-%m-%d %H:%M:%S", "{")
logger_handler.setFormatter(fmt=logger_formatter)
logger.addHandler(logger_handler)
FORMAT = '{asctime} - {levelname:8s} - {message}s'

from .communication import schema_version

__author__ = """RedRem95"""
__version__ = '0.5.0'
__schema_version__ = schema_version

logger.info(f"Loading j-chess-lib v-{__version__} by {__author__}. Using schema-version v-{schema_version}")



import logging
from queue import Queue
from typing import Optional
from uuid import UUID

from j_chess_lib.ai import StoreAI
from j_chess_lib.ai.container import GameState
from j_chess_lib.communication import MoveData, MatchStatusData, MatchFormatData

_logger = logging.getLogger("j_chess_lib")


class SampleAI(StoreAI):
    _COUNTER = 1

    def __init__(self):
        super(SampleAI, self).__init__(name=f"Sample-{self.__class__._COUNTER}")
        self.__class__._COUNTER += 1
        self._w_q = Queue()
        self._b_q = Queue()

    def new_match(self, match_id: UUID, enemy: str, match_format: MatchFormatData):
        super().new_match(match_id, enemy, match_format)
        _logger.info(f"{self} started a new match against {enemy}. Match is a {match_format.match_type_value}",
                     extra={"AI": self})

    def finalize_match(self, match_id: UUID, status: MatchStatusData, statistics: str):
        super().finalize_match(match_id, status, statistics)
        enemy, match_data = self.get_match(match_id=match_id)
        _logger.info(f"{self} finished the match against {enemy}", extra={"AI": self})
        _logger.info(f"{status.name_player1:10}: {status.score_player1}", extra={"AI": self})
        _logger.info(f"{status.name_player2:10}: {status.score_player2}", extra={"AI": self})

    def new_game(self, game_id: UUID, match_id: UUID, white_player: str):
        super().new_game(game_id, match_id, white_player)
        enemy, match_data = self.get_match(match_id=match_id)
        _logger.info(f"{self} started a new game against {enemy}. White is {white_player}", extra={"AI": self})
        self._w_q = Queue()
        self._b_q = Queue()
        self._w_q.put(MoveData(from_value="f2", to="f3"))
        self._w_q.put(MoveData(from_value="g2", to="g4"))
        self._b_q.put(MoveData(from_value="e7", to="e5"))
        self._b_q.put(MoveData(from_value="d8", to="h4"))

    def finalize_game(self, game_id: UUID, match_id: UUID, winner: Optional[str], pgn: str):
        super().finalize_game(game_id, match_id, winner, pgn)
        enemy, match_data = self.get_match(match_id=match_id)
        _logger.info(f"{self} finished the game against {enemy}. Winner is {winner}", extra={"AI": self})

    def get_move(self, game_id: UUID, match_id: UUID, game_state: GameState) -> MoveData:
        white_player = self.get_game(game_id=game_id, match_id=match_id)
        im_white = white_player == self.name
        _logger.info(f"{self} is expected to move. I am {'white' if im_white else 'black'}", extra={"AI": self})
        _logger.info(game_state.board_state.fen, extra={"AI": self})
        move_data = (self._w_q if im_white else self._b_q).get()
        return move_data



from j_chess_lib.communication.schema.j_chess_message import (
    AcceptMessage,
    AwaitMoveMessage,
    DisconnectMessage,
    DrawResponseMessage,
    ErrorType,
    GameOverMessage,
    GameStartMessage,
    HeartBeatMessage,
    JchessMessage,
    JchessMessageType,
    LoginMessage,
    LoginReplyMessage,
    MatchFormatData,
    MatchFoundMessage,
    MatchOverMessage,
    MatchStatusData,
    MatchStatusMessage,
    MatchTypeData,
    MatchTypeScore,
    MatchTypeValue,
    MatchTypeWinX,
    MoveData,
    MoveMessage,
    RequestDrawMessage,
    RequestDrawType,
    TimeControlData,
)

__all__ = [
    "AcceptMessage",
    "AwaitMoveMessage",
    "DisconnectMessage",
    "DrawResponseMessage",
    "ErrorType",
    "GameOverMessage",
    "GameStartMessage",
    "HeartBeatMessage",
    "JchessMessage",
    "JchessMessageType",
    "LoginMessage",
    "LoginReplyMessage",
    "MatchFormatData",
    "MatchFoundMessage",
    "MatchOverMessage",
    "MatchStatusData",
    "MatchStatusMessage",
    "MatchTypeData",
    "MatchTypeScore",
    "MatchTypeValue",
    "MatchTypeWinX",
    "MoveData",
    "MoveMessage",
    "RequestDrawMessage",
    "RequestDrawType",
    "TimeControlData",
]



from dataclasses import dataclass

from j_chess_lib.ai.board import BoardState
from j_chess_lib.communication.schema import AwaitMoveMessage, MoveData


@dataclass(frozen=True)
class GameState:
    enemy_time: int
    your_time: int
    last_move: MoveData
    board_state: BoardState

    @classmethod
    def from_await_move(cls, data: AwaitMoveMessage) -> "GameState":
        return GameState(
            enemy_time=data.time_control.enemy_time_in_ms,
            your_time=data.time_control.your_time_in_ms,
            last_move=data.last_move,
            board_state=BoardState(fen=data.position),
        )



from dataclasses import dataclass
from typing import Dict, Tuple, Optional, List


@dataclass(frozen=True)
class BoardState:
    fen: str

    def get_board(self) -> Dict[Tuple[str, int], Optional[str]]:
        ret: Dict[Tuple[str, int], Optional[str]] = {}
        row = 8
        col = ord('a')
        board_state = self.fen.strip().split(" ", maxsplit=1)[0].strip()
        for s in board_state:
            try:
                i = int(s)
                col += i
            except ValueError:
                if s == "/":
                    row -= 1
                    col = ord("a")
                else:
                    ret[(chr(col), row)] = s
                    col += 1
        return ret

    def string(self) -> str:
        b = self.get_board()
        ret: List[List[str]] = [[]]
        for row in range(8, 0, -1):
            for col in range(ord("a"), ord("h") + 1, 1):
                ret[-1].append(b.get((chr(col), row), " "))
            if row > 1:
                ret.append([])
        return "\n".join("".join(y for y in x) for x in ret)

    def white_turn(self):
        return self.fen.lower().strip().split(" ")[1] == "w"

    def __str__(self):
        return self.fen


if __name__ == '__main__':
    from j_chess_lib.ai.board.utilities import get_possible_moves, kill_king_move, is_promotion

    for fen in [
        # "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1",
        # "rnbqkbnr/pppppppp/8/8/4P3/8/PPPP1PPP/RNBQKBNR b KQkq e3 0 1",
        # "rnbqkbnr/pp1ppppp/8/3p4/4P3/8/PPPP1PPP/RNBQKBNR w KQkq c6 0 2",
        # "rnbqkbnr/pp1ppppp/8/3p4/4P3/5N2/PPPP1PPP/RNBQKB1R b KQkq - 1 2",
        # "rnbqkbnr/pp1ppppp/8/3p4/4P3/5N2/PPPP1PPP/RNBQKB1R w KQkq - 1 2",
        # "rnbqkbnr/1ppppppp/8/p7/P7/8/1PPPPPPP/RNBQKBNR w KQkq - 1 2",
        # "rnbqkbnr/1ppppppp/8/p3Q3/P7/k7/1PPPPPPP/RNBQKBNR w KQkq - 1 2",
        "8/4P3/8/8/8/3p4/4p3/8 b KQkq - 1 2",
    ]:
        board = BoardState(fen=fen)
        # pprint(board.get_board())
        print(f"\n{fen}:")
        print(f"  {' '.join(chr(x) for x in range(ord('a'), ord('h') + 1, 1))}")
        print(f" ┌{'─' * (8 + 7)}┐")
        for i, line in enumerate(board.string().split("\n")):
            print(f"{8 - i}│{' '.join(line)}│{8 - i}")
        print(f" └{'─' * (8 + 7)}┘")
        print(f"  {' '.join(chr(x) for x in range(ord('a'), ord('h') + 1, 1))}")
        print("")
        moves = get_possible_moves(board_state=board.get_board(), white=board.white_turn())
        print(len(moves), moves)
        print([x for x in moves if kill_king_move(board_state=board.get_board(), move=x)])
        print([x for x in moves if is_promotion(board_state=board.get_board(), move=x)])



import logging
import socket
from typing import Union

from xsdata.exceptions import ParserError
from xsdata.formats.dataclass.parsers import XmlParser
from xsdata.formats.dataclass.serializers import XmlSerializer
from xsdata.formats.dataclass.serializers.config import SerializerConfig

from .. import JchessMessage

_logger = logging.getLogger("j_chess_lib")


class ConnectionDecodeError(Exception):

    def __init__(self, message: str, raw_message: str):
        message = f"Client failed to decode received message{': ' if len(message) > 0 else ''}{message}"
        super().__init__(message)
        self._raw_message = raw_message

    @property
    def raw_message(self):
        return self._raw_message


class Connection:
    _PREFIX_BYTES = 4
    _ENDIAN_TYPE = "big"

    def __init__(self, address: str = "localhost", port: int = 5123):
        self._address = address
        self._port = port
        self._conn = socket.create_connection(address=(address, port))
        self._xml_parse = XmlParser()
        self._xml_serialize = XmlSerializer(config=SerializerConfig(pretty_print=False))
        self._send_count = 0
        self._recv_count = 0

    def __enter__(self):
        return self

    def disconnect(self):
        _logger.info("Disconnecting from connection")
        self._conn.close()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()

    def send(self, message: Union[JchessMessage, str]):
        if isinstance(message, str):
            length = len(message)
            # print(f"Sending message with length of {length}")
            length = length.to_bytes(self._PREFIX_BYTES, self._ENDIAN_TYPE)
            self._conn.send(length + message.encode('utf-8'))
            self._send_count += 1
            return
        if isinstance(message, JchessMessage):
            # print(f"Send message of type {message.message_type}")
            return self.send(message=self._xml_serialize.render(message))
        raise Exception(f"Can't send data of type {type(message)}")

    def recv(self) -> JchessMessage:
        message_byte_len = int.from_bytes(self._conn.recv(self._PREFIX_BYTES), self._ENDIAN_TYPE)
        self._recv_count += 1
        raw_message = self._conn.recv(message_byte_len).decode("utf-8")
        try:
            message = self._xml_parse.from_string(source=raw_message, clazz=JchessMessage)
        except ParserError as e:
            raise ConnectionDecodeError(message="", raw_message=raw_message) from e
        return message

    @property
    def send_count(self):
        return self._send_count

    @property
    def recv_count(self):
        return self._recv_count

    def __str__(self):
        return f"Connection to {self._address}:{self._port}"



try:
    from .schema import *
except ImportError as e:
    raise ImportError("Failed to import communication classes provided by schema. Did you install them?") from e

from .connection.Connection import Connection
schema_version = JchessMessage.schema_version

__all__ = [
    "Connection",
    "JchessMessage", "JchessMessageType", "schema_version", "MoveData", "MatchStatusData", "MatchFormatData"
]



from abc import ABC
from typing import Tuple

from ..communication import JchessMessage, ErrorType, JchessMessageType


class ClientError(Exception, ABC):
    pass


class UnhandledMessageError(ClientError):

    def __init__(self, message: JchessMessage):
        super(UnhandledMessageError, self).__init__(f"Message of hype {message.message_type} was unhandled")
        self._message = message


class LoginFailedError(ClientError):
    def __init__(self):
        super(LoginFailedError, self).__init__("Login Failed")


class WrongMessageType(ClientError):
    def __init__(self, message: JchessMessage, expected_type: Tuple[JchessMessageType, ...]):
        super(WrongMessageType, self).__init__(
            f"Expected message one of type {', '.join(str(x) for x in expected_type)} but got {message.message_type}"
        )
        self._message = message


class InterruptClient(Exception):
    def __init__(self, error: ErrorType):
        super(InterruptClient, self).__init__(f"Interruption called from error {error}")
        self._error = error

    @property
    def error(self) -> ErrorType:
        return self._error



from typing import Callable
from uuid import uuid4, UUID

from j_chess_lib.ai import AI
from j_chess_lib.ai.container import GameState
from j_chess_lib.communication import JchessMessage, JchessMessageType
from j_chess_lib.communication.schema import GameStartMessage, GameOverMessage, MoveMessage
from .Match import Match


class Game:
    def __init__(self, data: GameStartMessage, recv: Callable[[], JchessMessage], send: Callable[[JchessMessage], None],
                 ai: AI, match: Match):
        self._id = uuid4()
        self._recv = recv
        self._send = send
        self._data = data
        self._ai = ai
        self._match = match
        self._ai.new_game(game_id=self.id, match_id=match.id, white_player=self._data.name_white)

    @property
    def id(self) -> UUID:
        return self._id

    def play(self) -> GameOverMessage:

        while True:
            message = self._recv()
            if message.message_type == JchessMessageType.AWAIT_MOVE:
                move_data = self._ai.get_move(game_id=self.id, match_id=self._match.id,
                                              game_state=GameState.from_await_move(data=message.await_move))
                move_message = self._match.client.base_msg()
                move_message.message_type = JchessMessageType.MOVE
                move_message.move = MoveMessage(move=move_data)
                self._send(move_message)
            elif message.message_type == JchessMessageType.GAME_OVER:
                winner = None if message.game_over.is_draw else message.game_over.winner
                self._ai.finalize_game(game_id=self.id, match_id=self._match.id,
                                       winner=winner, pgn=message.game_over.pgn)
                return message.game_over



from abc import ABC, abstractmethod
from typing import Callable, List
from uuid import uuid4, UUID

from j_chess_lib.ai import AI
from j_chess_lib.communication import JchessMessage, JchessMessageType
from j_chess_lib.communication.schema import MatchFormatData, MatchFoundMessage, MatchOverMessage
from j_chess_lib.client.Exceptions import WrongMessageType
from j_chess_lib.client.Client import Client


class Match:

    def __init__(self, data: MatchFoundMessage, recv: Callable[[], JchessMessage],
                 send: Callable[[JchessMessage], None], ai: AI, client: Client):
        self._recv = recv
        self._send = send
        self._ai = ai
        self._client = client
        self._match_id: UUID = None
        self._enemy: str = None
        self._format: MatchFormatData = None
        self.new_match(data=data)

    @property
    def client(self) -> Client:
        return self._client

    @property
    def id(self) -> UUID:
        return self._match_id

    @classmethod
    def handle_match(cls, message: JchessMessage, recv: Callable[[], JchessMessage],
                     send: Callable[[JchessMessage], None], ai: AI, client: Client) -> "Match":
        if not message.message_type == JchessMessageType.MATCH_FOUND:
            raise WrongMessageType(message=message, expected_type=(JchessMessageType.MATCH_FOUND, ))

        return cls(data=message.match_found, recv=recv, send=send, ai=ai, client=client)

    def new_match(self, data: MatchFoundMessage):
        self._match_id = UUID(data.match_id)
        self._enemy = data.enemy_name
        self._format = data.match_format
        self._ai.new_match(match_id=self._match_id, enemy=self._enemy, match_format=self._format)

    def end_match(self, data: MatchOverMessage):
        self._ai.finalize_match(match_id=self._match_id, status=data.match_status, statistics=data.statistics)
        self._match_id = None
        self._enemy = None
        self._format = None

    def play_match(self):
        while True:
            message = self._recv()
            if message.message_type == JchessMessageType.MATCH_FOUND:
                self.new_match(data=message.match_found)
            elif message.message_type == JchessMessageType.MATCH_OVER:
                self.end_match(data=message.match_over)
            elif message.message_type == JchessMessageType.GAME_START:
                from .Game import Game
                game: Game = Game(data=message.game_start, recv=self._recv, send=self._send, ai=self._ai, match=self)
                message = game.play()






from typing import List, Tuple, Dict, Optional, Callable, Set


def _add(t1: Tuple[int, int], t2: Tuple[int, int]) -> Tuple[int, int]:
    return t1[0] + t2[0], t1[1] + t2[1]


def _mul(t1: Tuple[int, int], mul: int) -> Tuple[int, int]:
    return t1[0] * mul, t1[1] * mul


def get_possible_moves(
    board_state: Dict[Tuple[str, int], Optional[str]], white: bool
) -> List[Tuple[Tuple[str, int], Tuple[str, int]]]:
    ret: Set[Tuple[Tuple[str, int], Tuple[str, int]]] = set()

    fun: Callable[[str], bool] = str.isupper if white else str.islower
    for origin_position, figure in board_state.items():
        if not fun(figure):
            continue
        moves = get_next_position(origin_position=origin_position, board_state=board_state)
        ret.update([(origin_position, x[0]) for x in moves])
        # print(origin_position, figure)
        # print(moves)

    return list(ret)


def kill_king_move(board_state: Dict[Tuple[str, int], Optional[str]],
                   move: Tuple[Tuple[str, int], Tuple[str, int]]) -> bool:
    f, t = move
    if (f not in board_state) or (t not in board_state):
        return False
    white = board_state[f].isupper()
    enemy: Callable[[str], bool] = str.islower if white else str.isupper
    return enemy(board_state[t]) and board_state[t] in ("k", "K")


def in_chess(board_state: Dict[Tuple[str, int], Optional[str]], white: bool) -> bool:
    opponent_moves = get_possible_moves(board_state=board_state, white=not white)
    return any(kill_king_move(board_state=board_state, move=x) for x in opponent_moves)


def is_promotion(board_state: Dict[Tuple[str, int], Optional[str]],
                 move: Tuple[Tuple[str, int], Tuple[str, int]]) -> bool:
    f, t = move
    white = board_state[f].isupper()
    target_row = 8 if white else 1
    return board_state[f] in ("p", "P") and t[1] == target_row


def get_next_position(
    origin_position: Tuple[str, int], board_state: Dict[Tuple[str, int], Optional[str]]
) -> List[Tuple[Tuple[str, int], bool]]:
    if board_state[origin_position] in _FUNCTIONS:
        return _FUNCTIONS[board_state[origin_position]](origin_position, board_state)
    return []


def _get_next_position_straight(
    origin_position: Tuple[str, int], board_state: Dict[Tuple[str, int], Optional[str]], white: bool,
    directions: List[Tuple[int, int]], max_steps: int = 8
) -> List[Tuple[Tuple[str, int], bool]]:
    j = 1
    moves = []
    hits = []
    start_pos = (ord(origin_position[0]), origin_position[1])
    enemy: Callable[[str], bool] = str.islower if white else str.isupper
    a, h = ord("a"), ord("h")
    while j <= max_steps and len(directions) > 0:
        for i, direction in reversed(list(enumerate(directions))):
            direction: Tuple[int, int]
            new_pos_tmp = _add(start_pos, _mul(direction, j))
            if not 1 <= new_pos_tmp[1] <= 8 or not a <= new_pos_tmp[0] <= h:
                del directions[i]
                continue
            new_pos: Tuple[str, int] = (chr(new_pos_tmp[0]), new_pos_tmp[1])
            if new_pos in board_state:
                if enemy(board_state[new_pos]):
                    hits.append(new_pos)
                del directions[i]
                continue
            moves.append(new_pos)
        j += 1
    return [(x, False) for x in moves] + [(x, True) for x in hits]


def _get_next_position_rook(
    origin_position: Tuple[str, int], board_state: Dict[Tuple[str, int], Optional[str]], white: bool
) -> List[Tuple[Tuple[str, int], bool]]:
    directions = [(1, 0), (-1, 0), (0, 1), (0, -1)]
    return _get_next_position_straight(origin_position=origin_position, board_state=board_state, white=white,
                                       directions=directions)


def _get_next_position_knight(
    origin_position: Tuple[str, int], board_state: Dict[Tuple[str, int], Optional[str]], white: bool
) -> List[Tuple[Tuple[str, int], bool]]:
    hops = [(2, 1), (-2, 1), (2, -1), (-2, -1), (1, 2), (1, -2), (-1, 2), (-1, -2)]
    enemy: Callable[[str], bool] = str.islower if white else str.isupper
    start_pos = (ord(origin_position[0]), origin_position[1])
    a, h = ord("a"), ord("h")
    moves = [(chr(y[0]), y[1]) for y in (_add(x, start_pos) for x in hops) if a <= y[0] <= h and 1 <= y[1] <= 8]
    return [(x, x in board_state) for x in moves if x not in board_state or enemy(board_state[x])]


def _get_next_position_bishop(
    origin_position: Tuple[str, int], board_state: Dict[Tuple[str, int], Optional[str]], white: bool
) -> List[Tuple[Tuple[str, int], bool]]:
    directions = [(1, 1), (-1, 1), (1, -1), (-1, -1)]
    return _get_next_position_straight(origin_position=origin_position, board_state=board_state, white=white,
                                       directions=directions)


def _get_next_position_queen(
    origin_position: Tuple[str, int], board_state: Dict[Tuple[str, int], Optional[str]], white: bool
) -> List[Tuple[Tuple[str, int], bool]]:
    directions = [(1, 1), (-1, 1), (1, -1), (-1, -1), (1, 0), (-1, 0), (0, 1), (0, -1)]
    return _get_next_position_straight(origin_position=origin_position, board_state=board_state, white=white,
                                       directions=directions)


def _get_next_position_king(
    origin_position: Tuple[str, int], board_state: Dict[Tuple[str, int], Optional[str]], white: bool
) -> List[Tuple[Tuple[str, int], bool]]:
    directions = [(1, 1), (-1, 1), (1, -1), (-1, -1), (1, 0), (-1, 0), (0, 1), (0, -1)]
    return _get_next_position_straight(origin_position=origin_position, board_state=board_state, white=white,
                                       directions=directions, max_steps=1)


def _get_next_position_pawn(
    origin_position: Tuple[str, int], board_state: Dict[Tuple[str, int], Optional[str]], white: bool
) -> List[Tuple[Tuple[str, int], bool]]:
    direction = -1 + 2 * white
    home_row = 2 if white else 7
    moves = [(origin_position[0], origin_position[1] + direction)]
    if not 1 <= moves[0][1] <= 8:
        return []
    if origin_position[1] == home_row and moves[0] not in board_state:
        moves.append((origin_position[0], origin_position[1] + direction * 2))
    col = ord(origin_position[0])
    hits = [(chr(col + x), origin_position[1] + direction) for x in (-1, 1) if ord("a") <= col + x <= ord("h")]
    fun: Callable[[str], bool] = str.islower if white else str.isupper
    return [(x, False) for x in moves if x not in board_state] + \
           [(x, True) for x in hits if x in board_state and fun(board_state[x])]


_FUNCTIONS: Dict["str",
                 Callable[[Tuple[str, int], Dict[Tuple[str, int], Optional[str]]],
                          List[Tuple[Tuple[str, int], bool]]]] = {
    "p": lambda o, b: _get_next_position_pawn(origin_position=o, board_state=b, white=False),
    "P": lambda o, b: _get_next_position_pawn(origin_position=o, board_state=b, white=True),
    "r": lambda o, b: _get_next_position_rook(origin_position=o, board_state=b, white=False),
    "R": lambda o, b: _get_next_position_rook(origin_position=o, board_state=b, white=True),
    "n": lambda o, b: _get_next_position_knight(origin_position=o, board_state=b, white=False),
    "N": lambda o, b: _get_next_position_knight(origin_position=o, board_state=b, white=True),
    "b": lambda o, b: _get_next_position_bishop(origin_position=o, board_state=b, white=False),
    "B": lambda o, b: _get_next_position_bishop(origin_position=o, board_state=b, white=True),
    "q": lambda o, b: _get_next_position_queen(origin_position=o, board_state=b, white=False),
    "Q": lambda o, b: _get_next_position_queen(origin_position=o, board_state=b, white=True),
    "k": lambda o, b: _get_next_position_king(origin_position=o, board_state=b, white=False),
    "K": lambda o, b: _get_next_position_king(origin_position=o, board_state=b, white=True),
}



from dataclasses import dataclass, field
from enum import Enum
from typing import Optional


@dataclass
class DrawResponseMessage:
    accept: Optional[bool] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )


class ErrorType(Enum):
    NO_ERROR = "NO_ERROR"
    ERROR = "ERROR"
    AWAIT_LOGIN = "AWAIT_LOGIN"
    TIMEOUT = "TIMEOUT"
    TOO_MANY_TRIES = "TOO_MANY_TRIES"
    UNSUPPORTED_OPERATION = "UNSUPPORTED_OPERATION"
    DUPLICATE_NAME = "DUPLICATE_NAME"
    VERSION_MISMATCH = "VERSION_MISMATCH"


@dataclass
class GameOverMessage:
    winner: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )
    is_draw: Optional[bool] = field(
        default=None,
        metadata={
            "name": "isDraw",
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )
    pgn: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )


@dataclass
class GameStartMessage:
    name_white: Optional[str] = field(
        default=None,
        metadata={
            "name": "nameWhite",
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )


@dataclass
class HeartBeatMessage:
    pass


class JchessMessageType(Enum):
    ACCEPT = "Accept"
    LOGIN = "Login"
    LOGIN_REPLY = "LoginReply"
    HEART_BEAT = "HeartBeat"
    DISCONNECT = "Disconnect"
    MATCH_FOUND = "MatchFound"
    MATCH_OVER = "MatchOver"
    MATCH_STATUS = "MatchStatus"
    GAME_START = "GameStart"
    GAME_OVER = "GameOver"
    AWAIT_MOVE = "AwaitMove"
    MOVE = "Move"
    REQUEST_DRAW = "RequestDraw"
    DRAW_RESPONSE = "DrawResponse"


@dataclass
class LoginMessage:
    name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )


@dataclass
class LoginReplyMessage:
    new_id: Optional[str] = field(
        default=None,
        metadata={
            "name": "newId",
            "type": "Element",
            "namespace": "",
            "required": True,
            "pattern": r"([0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12})|(\{[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}\})",
        }
    )


@dataclass
class MatchStatusData:
    name_player1: Optional[str] = field(
        default=None,
        metadata={
            "name": "namePlayer1",
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )
    name_player2: Optional[str] = field(
        default=None,
        metadata={
            "name": "namePlayer2",
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )
    score_player1: Optional[int] = field(
        default=None,
        metadata={
            "name": "scorePlayer1",
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )
    score_player2: Optional[int] = field(
        default=None,
        metadata={
            "name": "scorePlayer2",
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )


@dataclass
class MatchTypeData:
    pass


class MatchTypeValue(Enum):
    WIN_X = "WIN_X"
    SCORE = "SCORE"


@dataclass
class MoveData:
    from_value: Optional[str] = field(
        default=None,
        metadata={
            "name": "from",
            "type": "Element",
            "namespace": "",
            "required": True,
            "pattern": r"[abcdefgh][1-8]",
        }
    )
    to: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
            "pattern": r"[abcdefgh][1-8]",
        }
    )
    promotion_unit: Optional[str] = field(
        default=None,
        metadata={
            "name": "promotionUnit",
            "type": "Element",
            "namespace": "",
            "pattern": r"[nNbBrRqQ]",
        }
    )


class RequestDrawType(Enum):
    OFFER = "OFFER"
    FIFTY_MOVE_RULE = "FIFTY_MOVE_RULE"


@dataclass
class TimeControlData:
    your_time_in_ms: Optional[int] = field(
        default=None,
        metadata={
            "name": "yourTimeInMs",
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )
    enemy_time_in_ms: Optional[int] = field(
        default=None,
        metadata={
            "name": "enemyTimeInMs",
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )


@dataclass
class AcceptMessage:
    accept: Optional[bool] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )
    error_type_code: Optional[ErrorType] = field(
        default=None,
        metadata={
            "name": "errorTypeCode",
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )


@dataclass
class AwaitMoveMessage:
    position: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
            "pattern": r"\s*([rnbqkpRNBQKP1-8]+/){7}([rnbqkpRNBQKP1-8]+)\s[bw-]\s(([kqKQ]{1,4})|(-))\s(([a-h][36])|(-))\s\d+\s\d+\s*",
        }
    )
    last_move: Optional[MoveData] = field(
        default=None,
        metadata={
            "name": "lastMove",
            "type": "Element",
            "namespace": "",
        }
    )
    time_control: Optional[TimeControlData] = field(
        default=None,
        metadata={
            "name": "timeControl",
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )


@dataclass
class DisconnectMessage:
    error_type_code: Optional[ErrorType] = field(
        default=None,
        metadata={
            "name": "errorTypeCode",
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )


@dataclass
class MatchFormatData:
    match_type_value: Optional[MatchTypeValue] = field(
        default=None,
        metadata={
            "name": "matchTypeValue",
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )
    match_type_data: Optional[MatchTypeData] = field(
        default=None,
        metadata={
            "name": "matchTypeData",
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )
    time_per_side: Optional[int] = field(
        default=None,
        metadata={
            "name": "timePerSide",
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )
    time_per_side_increment: Optional[int] = field(
        default=None,
        metadata={
            "name": "timePerSideIncrement",
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )
    time_per_side_per_move: Optional[int] = field(
        default=None,
        metadata={
            "name": "timePerSidePerMove",
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )


@dataclass
class MatchStatusMessage:
    match_status: Optional[MatchStatusData] = field(
        default=None,
        metadata={
            "name": "matchStatus",
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )


@dataclass
class MatchTypeScore(MatchTypeData):
    amount_to_play: Optional[int] = field(
        default=None,
        metadata={
            "name": "amountToPlay",
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )
    points_per_win: int = field(
        init=False,
        default=2,
        metadata={
            "name": "pointsPerWin",
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )
    points_per_draw: int = field(
        init=False,
        default=1,
        metadata={
            "name": "pointsPerDraw",
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )


@dataclass
class MatchTypeWinX(MatchTypeData):
    amount_to_win: Optional[int] = field(
        default=None,
        metadata={
            "name": "amountToWin",
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )


@dataclass
class MoveMessage:
    move: Optional[MoveData] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )


@dataclass
class RequestDrawMessage:
    reason: Optional[RequestDrawType] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )


@dataclass
class MatchFoundMessage:
    match_id: Optional[str] = field(
        default=None,
        metadata={
            "name": "matchId",
            "type": "Element",
            "namespace": "",
            "required": True,
            "pattern": r"([0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12})|(\{[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}\})",
        }
    )
    enemy_name: Optional[str] = field(
        default=None,
        metadata={
            "name": "enemyName",
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )
    match_format: Optional[MatchFormatData] = field(
        default=None,
        metadata={
            "name": "matchFormat",
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )


@dataclass
class MatchOverMessage:
    match_status: Optional[MatchStatusData] = field(
        default=None,
        metadata={
            "name": "matchStatus",
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )
    match_format: Optional[MatchFormatData] = field(
        default=None,
        metadata={
            "name": "matchFormat",
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )
    statistics: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        }
    )


@dataclass
class JchessMessage:
    class Meta:
        name = "JChessMessage"

    accept: Optional[AcceptMessage] = field(
        default=None,
        metadata={
            "name": "Accept",
            "type": "Element",
            "namespace": "",
        }
    )
    login: Optional[LoginMessage] = field(
        default=None,
        metadata={
            "name": "Login",
            "type": "Element",
            "namespace": "",
        }
    )
    login_reply: Optional[LoginReplyMessage] = field(
        default=None,
        metadata={
            "name": "LoginReply",
            "type": "Element",
            "namespace": "",
        }
    )
    heart_beat: Optional[HeartBeatMessage] = field(
        default=None,
        metadata={
            "name": "HeartBeat",
            "type": "Element",
            "namespace": "",
        }
    )
    disconnect: Optional[DisconnectMessage] = field(
        default=None,
        metadata={
            "name": "Disconnect",
            "type": "Element",
            "namespace": "",
        }
    )
    match_found: Optional[MatchFoundMessage] = field(
        default=None,
        metadata={
            "name": "MatchFound",
            "type": "Element",
            "namespace": "",
        }
    )
    match_over: Optional[MatchOverMessage] = field(
        default=None,
        metadata={
            "name": "MatchOver",
            "type": "Element",
            "namespace": "",
        }
    )
    match_status: Optional[MatchStatusMessage] = field(
        default=None,
        metadata={
            "name": "MatchStatus",
            "type": "Element",
            "namespace": "",
        }
    )
    game_start: Optional[GameStartMessage] = field(
        default=None,
        metadata={
            "name": "GameStart",
            "type": "Element",
            "namespace": "",
        }
    )
    game_over: Optional[GameOverMessage] = field(
        default=None,
        metadata={
            "name": "GameOver",
            "type": "Element",
            "namespace": "",
        }
    )
    await_move: Optional[AwaitMoveMessage] = field(
        default=None,
        metadata={
            "name": "AwaitMove",
            "type": "Element",
            "namespace": "",
        }
    )
    move: Optional[MoveMessage] = field(
        default=None,
        metadata={
            "name": "Move",
            "type": "Element",
            "namespace": "",
        }
    )
    request_draw: Optional[RequestDrawMessage] = field(
        default=None,
        metadata={
            "name": "RequestDraw",
            "type": "Element",
            "namespace": "",
        }
    )
    draw_response: Optional[DrawResponseMessage] = field(
        default=None,
        metadata={
            "name": "DrawResponse",
            "type": "Element",
            "namespace": "",
        }
    )
    message_type: Optional[JchessMessageType] = field(
        default=None,
        metadata={
            "name": "messageType",
            "type": "Attribute",
            "required": True,
        }
    )
    player_id: Optional[str] = field(
        default=None,
        metadata={
            "name": "playerId",
            "type": "Attribute",
            "required": True,
            "pattern": r"([0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12})|(\{[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}\})",
        }
    )
    schema_version: str = field(
        init=False,
        default="0.1.1",
        metadata={
            "name": "schemaVersion",
            "type": "Attribute",
            "required": True,
        }
    )



#!/usr/bin/env python

"""The setup script."""
import os
from distutils.command import build_py, install_lib, sdist, bdist, upload
from distutils.dist import Distribution
from typing import Any, Type

from setuptools import setup, find_packages, Command

schema_package = os.environ.get("j-chess-schema-package", "j_chess_lib.communication.schema")
os.makedirs(schema_package.replace(".", "/"), exist_ok=True)

_installed_schema = [False]


def install_schema_package():
    if not _installed_schema[0]:
        print("-" * 10, "schema install", "-" * 10)
        xsd_url = os.environ.get("j-chess-xsd-url",
                                 "https://raw.githubusercontent.com/JoKrus/j-chess-xsd/master/jChessMessage.xsd")
        import subprocess
        try:
            subprocess.run(["rm", "-rf", "j_chess_lib/communication/schema"], check=True)
            print("Removed old schema data")
        except subprocess.CalledProcessError:
            pass
        print("Load schema data from %s and install them" % xsd_url)
        subprocess.run(["xsdata", xsd_url, "--package", schema_package], check=True)
        print("Installed new schema data")
        print("-" * 10, "schema install", "-" * 10)
    else:
        print("-" * 10, "schema already installed", "-" * 10)
    _installed_schema[0] = True


class InstallSchema(Command):
    key = "install_schema"
    description = "Create classes from the j-chess schema"
    user_options = []

    def __init__(self, dist: Distribution, **kw: Any):
        super().__init__(dist, **kw)

    def finalize_options(self) -> None:
        pass

    def initialize_options(self) -> None:
        pass

    def run(self):
        install_schema_package()


def install_schema_wrapper(class_type) -> Type[Command]:
    class Wrapper(class_type):

        def run(self) -> None:
            install_schema_package()
            super(Wrapper, self).run()

    return Wrapper


cmd_classes = {
    InstallSchema.key: InstallSchema,
    "build_py": install_schema_wrapper(build_py.build_py),
    "install_lib": install_schema_wrapper(install_lib.install_lib),
    "sdist": install_schema_wrapper(sdist.sdist),
    "bdist": install_schema_wrapper(bdist.bdist),
    "upload": install_schema_wrapper(upload.upload),
}

try:
    from wheel.bdist_wheel import bdist_wheel

    cmd_classes["bdist_wheel"] = install_schema_wrapper(bdist_wheel)
except ImportError:
    pass

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = ["xsdata[cli,lxml,soap]"]
test_requirements = []

setup(
    author="RedRem95",
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    description="Python library for a j-chess bot. Beep Boop",
    install_requires=requirements,
    license="GNU General Public License v3",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='j_chess_lib',
    name='j_chess_lib',
    packages=find_packages(include=['j_chess_lib', 'j_chess_lib.*']) + [schema_package],
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/RedRem95/j_chess_lib',
    version='0.5.0',
    zip_safe=True,
    cmdclass=cmd_classes,
)



from abc import ABC, abstractmethod
from typing import Optional, Tuple, Dict, Any, List
from uuid import UUID

from .container import GameState
from ..communication.schema import MatchStatusData, MatchFormatData, MoveData


class AI(ABC):

    def __init__(self, name: str = None):
        self._name = f"{self.__class__.__name__}-PyAI" if name is None else name

    @property
    def name(self) -> str:
        return self._name

    def __str__(self):
        return self.name

    @abstractmethod
    def new_match(self, match_id: UUID, enemy: str, match_format: MatchFormatData):
        pass

    @abstractmethod
    def finalize_match(self, match_id: UUID, status: MatchStatusData, statistics: str):
        pass

    @abstractmethod
    def new_game(self, game_id: UUID, match_id: UUID, white_player: str):
        pass

    @abstractmethod
    def finalize_game(self, game_id: UUID, match_id: UUID, winner: Optional[str], pgn: str):
        pass

    @abstractmethod
    def get_move(self, game_id: UUID, match_id: UUID, game_state: GameState) -> MoveData:
        pass

    # noinspection PyMethodMayBeStatic
    def metrics(self) -> List[Tuple[str, Any]]:
        return [("name", self.name)]


class DumbAI(AI, ABC):

    def new_match(self, match_id: UUID, enemy: str, match_format: MatchFormatData):
        pass

    def finalize_match(self, match_id: UUID, status: MatchStatusData, statistics: str):
        pass

    def new_game(self, game_id: UUID, match_id: UUID, white_player: str):
        pass

    def finalize_game(self, game_id: UUID, match_id: UUID, winner: Optional[str], pgn: str):
        pass


class StoreAI(AI, ABC):

    def __init__(self, name: str = None):
        super().__init__(name)
        self._match_storage: Dict[UUID, Tuple[str, MatchFormatData]] = {}
        self._game_storage: Dict[Tuple[UUID, UUID], str] = {}
        self._current_match: Optional[UUID] = None
        self._current_game: Optional[Tuple[UUID, UUID]] = None

    def get_match(self, match_id: UUID) -> Optional[Tuple[str, MatchFormatData]]:
        return self._match_storage.get(match_id, None)

    def get_game(self, game_id: UUID, match_id: UUID) -> Optional[str]:
        return self._game_storage.get((match_id, game_id), None)

    def new_match(self, match_id: UUID, enemy: str, match_format: MatchFormatData):
        self._match_storage[match_id] = (enemy, match_format)
        self._current_match = match_id

    def finalize_match(self, match_id: UUID, status: MatchStatusData, statistics: str):
        pass

    def new_game(self, game_id: UUID, match_id: UUID, white_player: str):
        self._game_storage[(match_id, game_id)] = white_player
        self._current_game = (game_id, match_id)

    def finalize_game(self, game_id: UUID, match_id: UUID, winner: Optional[str], pgn: str):
        pass

    def metrics(self) -> List[Tuple[str, Any]]:
        metrics = super().metrics()
        if self._current_match is not None:
            enemy, match = self.get_match(match_id=self._current_match)
            match: MatchFormatData
            metrics.append(("In match", self._current_match))
            metrics.append(("   against", enemy))
            metrics.append(("   match type", f"{match.match_type_data}"))
        if self._current_game is not None:
            game_id, match_id = self._current_game
            enemy, match = self.get_match(match_id=match_id)
            white_player = self.get_game(game_id=game_id, match_id=match_id)
            metrics.append(("In game", game_id))
            metrics.append(("   of match", f"{match_id} -> {match.match_type_value.value}"))
            metrics.append(("   against", enemy))
            metrics.append(("   white player", f"{white_player} <- {'me' if white_player == self.name else 'enemy'}"))
        return metrics
